import bisect
import functools
import time
import inspect
from typing import Any
from abc import ABC, abstractmethod
from gtoolkit_bridge import gtView

def methodevent(message):
    def decorate(func):
        @functools.wraps(func)
        def wrapped_function(*args, **kwargs):
            if 'signals' not in globals():
                return func(*args, **kwargs)
            signal = MethodStartSignal(message)
            signal.file = inspect.getsourcefile(func)
            [_, signal.line] = inspect.getsourcelines(func)
            try:
                value = func(*args, **kwargs)
                return value
            finally:
                MethodEndSignal(message)
        return wrapped_function
    return decorate

def argmethodevent(message):
    def decorate(func):
        @functools.wraps(func)
        def wrapped_function(*args, **kwargs):
            if 'signals' not in globals():
                return func(*args, **kwargs)           
            signal = ArgumentMethodStartSignal(message, kwargs)
            signal.file = inspect.getsourcefile(func)
            [_, signal.line] = inspect.getsourcelines(func)
            try:
                value = func(*args, **kwargs)
                return value
            finally:
                MethodEndSignal(message)
        return wrapped_function
    return decorate

class Telemetry(ABC):
    def __init__(self, message):
        super().__init__
        self.message = message

    def children(self):
        return []

    @abstractmethod
    def timestamp(self):
        pass

    @abstractmethod
    def duration(self):
        pass

class TelemetrySignal(Telemetry):
    def __init__(self, message) -> None:
        super().__init__(message)
        self._timestamp = time.perf_counter_ns()
        cf = inspect.stack()[1]
        self.file = cf.filename
        self.line = cf.lineno

        if 'signals' in globals():
            global signals
            signals.add_signal(self)

    def timestamp(self):
        return self._timestamp

    def duration(self):
        return 0

    def isStartSignal(self):
        return False
    
    def isEndSignal(self):
        return False
    
    @gtView
    def gtViewSignalTree(self, aBuilder):
        return aBuilder.columnedTree()\
            .title("Tree")\
            .priority(2)\
            .items(lambda: [self])\
            .children(lambda each: each.children())\
            .column("Message", lambda each: each.message)\
            .column("Duration", lambda each: each.duration())

class TelemetryEvent(Telemetry):
    def __init__(self, message):
        super().__init__(message)
        self._children = []
        self.startSignal = None
        self.endSignal = None

    def children(self):
        return self._children

    def addChild(self, child):
        bisect.insort(self._children, child, key=lambda x:x.timestamp())

    def timestamp(self):
        return self.startSignal.timestamp()

    def duration(self):
        return self.endSignal.timestamp() - self.startSignal.timestamp()
    
    @gtView
    def gtViewEventTree(self, aBuilder):
        return aBuilder.columnedTree()\
            .title("Tree")\
            .priority(2)\
            .items(lambda: [ self ])\
            .children(lambda each: each.children())\
            .column("Message", lambda each: each.message)\
            .column("Duration", lambda each: each.duration())

class MethodStartSignal(TelemetrySignal):
    def isStartSignal(self):
        return True
    
    def isEndSignal(self):
        return False

class MethodEndSignal(TelemetrySignal):
    def isStartSignal(self):
        return False
    
    def isEndSignal(self):
        return True

class ArgumentMethodStartSignal(MethodStartSignal):
    def __init__(self, message, args):
        super().__init__(message)
        self.args = args.copy()

class TelemetrySignalGroup:
    def __init__(self) -> None:
        self.signals = []

    def get_signals(self):
        return self.signals
    
    def get_event_tree(self):
        b = self.get_signals()
        value = []
        index = 0
        while index < len(b):
            [index, tree] = self.compute_tree(index, b, 0)
            value.append(tree)
        return value
    
    def compute_tree(self, index, list, depth):
        print(f"Depth: {depth}:{index}/{len(list)}")
        if index >= len(list):
            return [index, []]
        if not list[index].isStartSignal():
            return [index+1, list[index]]    # leaf signals
        root = TelemetryEvent(list[index].message)
        root.startSignal = list[index]
        index = index + 1
        while index < len(list) and not list[index].isEndSignal():
            [newindex, kid] = self.compute_tree(index, list, depth+1)
            root.addChild(kid)
            index = newindex
        if index < len(list):
            root.endSignal = list[index]
            index = index + 1
        return [index, root]

    def add_signal(self, signal):
        self.signals.append(signal)

    @gtView
    def gtViewSignals(self, aBuilder):
        return aBuilder.columnedList()\
            .title("Signals")\
            .priority(1)\
            .items(lambda: self.get_signals())\
            .column("Signal Class", lambda each:f"{each.__class__.__name__}")\
            .column("Message", lambda each: each.message)\
            .column("Timestamp", lambda each: each.timestamp())
    
    @gtView
    def gtViewSignalTree(self, aBuilder):
        return aBuilder.columnedTree()\
            .title("Tree")\
            .priority(2)\
            .items(lambda: self.get_event_tree())\
            .children(lambda each: each.children())\
            .column("Message", lambda each: each.message)\
            .column("Duration", lambda each: each.duration())

def start_signals():
    reset_signals()

def reset_signals():
    global signals
    signals = TelemetrySignalGroup()

def get_signals():
    global signals
    return signals


class ListColumn:
	def __init__(self, title, formatCallback, width):
		self.title = title
		self.formatCallback = formatCallback
		self.width = width

	def getTitle(self):
		return self.title
	
	def formatItem(self, item):
		return self.formatCallback(item)


from .view import View

class ForwardView(View):
	def __init__(self):
		super().__init__()
		self.objectComputation = lambda : None
		self.forwardView = None
		self.computedView = None

	def object(self, objectComputation):
		self.objectComputation = objectComputation
		return self
	
	def view(self, viewName):
		self.forwardView = viewName
		return self

	def getForwardObject(self):
		return self.objectComputation()

	def getForwardView(self):
		return self.forwardView

	def dataSource(self):
		return self

	def getDataSource(self, viewName):
		return self.computedView.dataSource()

	def getViewSpecificationForForwarding(self):
		from .view_builder import ViewBuilder
		self.computedView = getattr(self.getForwardObject(), self.getForwardView())(ViewBuilder())
		exportData = self.computedView.asDictionaryForExport()
		exportData["methodSelector"] = self.getForwardView()
		return exportData

	def asDictionaryForExport(self):
		exportData = super().asDictionaryForExport()
		exportData["viewName"] = "GtPhlowForwardViewSpecification"
		exportData["dataTransport"] = 2
		return exportData


from uuid import uuid1
from .bridge_globals import bridge_globals

def ensure_global_registry():
    if not hasattr(bridge_globals(), 'obj_registry'):
        bridge_globals()['obj_registry'] = Registry()

def registry():
    return bridge_globals()['obj_registry']

class Registry():

    def __init__(self):
        self.idToObjMap = {}
        self.objToIdMap = {}

    def hasId(self, anId):
        return anId in self.idToObjMap

    def createNewObjId(self):
        return uuid1().hex
    
    def register(self, obj):
        if obj is None:
            return 0
        if id(obj) in self.objToIdMap:
            return self.objToIdMap[id(obj)]
        else:
            return self._register(obj, self.createNewObjId())
    
    def register_with_id(self, obj, newObjId):
        if obj is None:
            return RegisterForbiddenObject(obj)
        if id(obj) in self.objToIdMap:
            objId = self.objToIdMap[id(obj)]
            if objId == newObjId:
                return newObjId
            else:
                raise RegisterWithDifferentIdError(obj, newObjId)
        else:
            return self._register(obj, newObjId)

    def resolve(self, objId):
        if objId in self.idToObjMap:
            return self.idToObjMap[objId]
        else:
            raise ResolveUnknownObject(objId)

    def _register(self, obj, newObjId):
        self.idToObjMap[newObjId] = obj
        self.objToIdMap[id(obj)] = newObjId
        return newObjId

    def clean(self, objId):
        obj = self.idToObjMap[objId]
        del self.idToObjMap[objId]
        del self.objToIdMap[id(obj)]

    def isProxy(self, anObject):
        is_proxy = False

        if isinstance(anObject, dict):
            if len(anObject.keys()) == 2 and ('__pyclass__' in anObject) and ('__pyid__' in anObject):
                is_proxy = True

        return is_proxy

    def proxy(self, obj):
        if obj is None:
            return obj
        return {
            '__pyclass__': type(obj).__name__,
            '__pyid__': self.register(obj)
            }

class RegistryError(Exception):
    pass

class RegisterWithDifferentIdError(RegistryError):
    def __init__(self, obj, newId):
        RegistryError.__init__(self,"Attempting to register object {0} with ID {1} with different ID {2}.".format(type(obj).__name__, registry().register(obj), newId))

class ResolveUnknownObject(RegistryError):
    def __init__(self, objId):
        RegistryError.__init__(self,"Attempting to resolve unknown object with id {0}.".format(objId))

class RegisterForbiddenObject(RegistryError):
    def __init__(self, obj):
        RegistryError.__init__(self,"Attempting to register forbidden object of type {0}.".format(type(obj).__name__))

ensure_global_registry()



from gtoolkit_bridge import gtView

class GtRemotePhlowDeclarativeTestInspectable:

	def __init__(self):
		super().__init__()
		self.collection = [42, "Lorem ipsum dolor sit amet", 3.14159, True, False, None]
		self.string = "Hello GT World !"

	def get_collection(self):
		return self.collection

	def get_string(self):
		return self.string

	@gtView
	def gtViewCollection(self, builder):
		clist = builder.columnedList()
		clist.title('Collection')
		clist.priority(25)
		clist.items(lambda: list(range(0, len(self.get_collection()))))
		clist.column('#', lambda index: index)
		clist.column('item', lambda index: str(self.get_collection()[index]))
		clist.set_accessor(lambda index: self.get_collection()[index])
		return clist

	@gtView
	def gtViewString(self, builder):
		editor = builder.textEditor()
		editor.title("String")
		editor.priority(35)
		editor.setString(self.get_string())
		return editor

	@gtView
	def gtViewError(self, builder):
		raise Exception("I'm sorry, Dave. I'm afraid I can't do that.")
		
	@gtView
	def gtViewEmpty(self, builder):
		return builder.empty()
		
	@gtView
	def gtViewCollectionForward(self, builder):
		forward = builder.forward()
		forward.title("Collection forward")
		forward.priority(30)
		forward.object(lambda: self)
		forward.view('gtViewCollection')
		return forward
